# Copyright Coralogix Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from copy import deepcopy

import boto3
import pytest
from botocore.exceptions import ClientError

from tests.bedrock.utils import (
    IMAGE_DATA,
    assert_attributes_in_span,
    assert_expected_metrics,
    assert_tool_definitions_in_span,
)
from tests.utils import assert_choices_in_span, assert_messages_in_span


def _get_current_weather_tool_definition():
    return {
        "toolSpec": {
            "name": "get_current_weather",
            "description": "Get the current weather in a given location",
            "inputSchema": {
                "json": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city and state, e.g. Boston, MA",
                        },
                    },
                    "required": ["location"],
                    "additionalProperties": False,
                },
            },
        },
    }


def _convert_stream_to_response(stream) -> dict:
    result = {
        "stopReason": "",
        "output": {
            "usage": {},
            "message": {
                "content": [],
            },
        },
    }
    current_block = {}
    for event in stream:
        if "messageStart" in event:
            result["output"]["message"]["role"] = event["messageStart"]["role"]
        if "contentBlockStart" in event:
            start = event["contentBlockStart"].get("start", {})
            if "toolUse" in start:
                current_block = {"toolUse": deepcopy(start["toolUse"])}
        if "contentBlockDelta" in event:
            delta = event["contentBlockDelta"].get("delta", {})
            if "text" in delta:
                current_block.setdefault("text", "")
                current_block["text"] += delta["text"]
            elif "toolUse" in delta:
                current_block["toolUse"].setdefault("input", "")
                current_block["toolUse"]["input"] += delta["toolUse"]["input"]
        if "contentBlockStop" in event:
            if "toolUse" in current_block:
                current_block["toolUse"]["input"] = json.loads(
                    current_block["toolUse"]["input"]
                )
            result["output"]["message"]["content"].append(current_block)
            current_block = {}
        if "metadata" in event:
            result["usage"] = event["metadata"]["usage"]
        if "messageStop" in event:
            result["stopReason"] = event["messageStop"]["stopReason"]

    return result


def _run_and_check_converse(
    bedrock_client,
    model_id: str,
    span_exporter,
    metric_reader,
    expect_content: bool,
    stream: bool,
):
    args = {
        "modelId": model_id,
        "system": [{"text": "you are a helpful assistant"}],
        "messages": [{"role": "user", "content": [{"text": "say this is a test"}]}],
        "inferenceConfig": {
            "maxTokens": 300,
            "temperature": 0,
            "topP": 1,
        },
    }
    if stream:
        span_name = "bedrock.converse_stream"
        stream_result = bedrock_client.converse_stream(**args)
        result = _convert_stream_to_response(stream_result["stream"])
    else:
        span_name = "bedrock.converse"
        result = bedrock_client.converse(**args)

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    assert_attributes_in_span(
        span=spans[0],
        span_name=span_name,
        request_model=model_id,
        response_model=model_id,
        usage_input_tokens=result["usage"]["inputTokens"],
        usage_output_tokens=result["usage"]["outputTokens"],
        finish_reasons=(result["stopReason"],),
        max_tokens=300,
        temperature=0,
        top_p=1,
    )

    expected_messages = [
        {"role": "system", "content": "you are a helpful assistant"},
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0],
        expected_messages=expected_messages,
        expect_content=expect_content,
    )

    expected_choice = {
        "finish_reason": result["stopReason"],
        "message": {
            "role": result["output"]["message"]["role"],
            "content": result["output"]["message"]["content"][0]["text"],
        },
    }
    assert_choices_in_span(
        span=spans[0], expected_choices=[expected_choice], expect_content=expect_content
    )

    metrics = metric_reader.get_metrics_data().resource_metrics
    assert len(metrics) == 1

    metric_data = metrics[0].scope_metrics[0].metrics
    assert_expected_metrics(
        metrics=metric_data,
        request_model=model_id,
        response_model=model_id,
        usage_input_tokens=result["usage"]["inputTokens"],
        usage_output_tokens=result["usage"]["outputTokens"],
    )


def _run_and_check_converse_tool_calls(
    bedrock_client,
    model_id: str,
    span_exporter,
    metric_reader,
    expect_content: bool,
    stream: bool,
):
    tool_definition = _get_current_weather_tool_definition()
    args = {
        "modelId": model_id,
        "system": [{"text": "you are a helpful assistant"}],
        "messages": [
            {
                "role": "user",
                "content": [
                    {"text": "What's the weather in Seattle and San Francisco today?"}
                ],
            }
        ],
        "toolConfig": {"tools": [tool_definition]},
    }
    if stream:
        span_name = "bedrock.converse_stream"
        stream_result_0 = bedrock_client.converse_stream(**args)
        result_0 = _convert_stream_to_response(stream_result_0["stream"])
    else:
        span_name = "bedrock.converse"
        result_0 = bedrock_client.converse(**args)
    tool_call_blocks = [
        block
        for block in result_0["output"]["message"]["content"]
        if "toolUse" in block
    ]

    assert result_0["stopReason"] == "tool_use"

    tool_results = [
        {
            "tool_call_id": tool_call_blocks[0]["toolUse"]["toolUseId"],
            "content": "50 degrees and raining",
        },
        {
            "tool_call_id": tool_call_blocks[1]["toolUse"]["toolUseId"],
            "content": "70 degrees and sunny",
        },
    ]

    args["messages"].append(result_0["output"]["message"])
    args["messages"].append(
        {
            "role": "user",
            "content": [
                {
                    "toolResult": {
                        "toolUseId": tool_results[0]["tool_call_id"],
                        "content": [{"text": tool_results[0]["content"]}],
                    }
                },
                {
                    "toolResult": {
                        "toolUseId": tool_results[1]["tool_call_id"],
                        "content": [{"text": tool_results[1]["content"]}],
                    }
                },
            ],
        }
    )

    if stream:
        stream_result_1 = bedrock_client.converse_stream(**args)
        result_1 = _convert_stream_to_response(stream_result_1["stream"])
    else:
        result_1 = bedrock_client.converse(**args)
    assert result_1["stopReason"] == "end_turn"

    expected_assistant_message = {
        "role": result_0["output"]["message"]["role"],
        "content": result_0["output"]["message"]["content"][0]["text"],
        "tool_calls": [
            {
                "id": tool_call_blocks[0]["toolUse"]["toolUseId"],
                "type": "function",
                "function": {
                    "name": tool_call_blocks[0]["toolUse"]["name"],
                    "arguments": json.dumps(tool_call_blocks[0]["toolUse"]["input"]),
                },
            },
            {
                "id": tool_call_blocks[1]["toolUse"]["toolUseId"],
                "type": "function",
                "function": {
                    "name": tool_call_blocks[1]["toolUse"]["name"],
                    "arguments": json.dumps(tool_call_blocks[1]["toolUse"]["input"]),
                },
            },
        ],
    }

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 2

    assert_attributes_in_span(
        span=spans[0],
        span_name=span_name,
        request_model=model_id,
        response_model=model_id,
        usage_input_tokens=result_0["usage"]["inputTokens"],
        usage_output_tokens=result_0["usage"]["outputTokens"],
        finish_reasons=(result_0["stopReason"],),
    )
    assert_attributes_in_span(
        span=spans[1],
        span_name=span_name,
        request_model=model_id,
        response_model=model_id,
        usage_input_tokens=result_1["usage"]["inputTokens"],
        usage_output_tokens=result_1["usage"]["outputTokens"],
        finish_reasons=(result_1["stopReason"],),
    )

    expected_tool_definition = {
        "name": tool_definition["toolSpec"]["name"],
        "description": tool_definition["toolSpec"]["description"],
        "parameters": json.dumps(tool_definition["toolSpec"]["inputSchema"]["json"]),
    }
    assert_tool_definitions_in_span(spans[0], [expected_tool_definition])
    assert_tool_definitions_in_span(spans[1], [expected_tool_definition])

    expected_messages_0 = [
        {"role": "system", "content": "you are a helpful assistant"},
        {
            "role": "user",
            "content": "What's the weather in Seattle and San Francisco today?",
        },
    ]
    assert_messages_in_span(
        span=spans[0],
        expected_messages=expected_messages_0,
        expect_content=expect_content,
    )

    expected_messages_1 = [
        *expected_messages_0,
        expected_assistant_message,
        {"role": "tool", **tool_results[0]},
        {"role": "tool", **tool_results[1]},
    ]
    assert_messages_in_span(
        span=spans[1],
        expected_messages=expected_messages_1,
        expect_content=expect_content,
    )

    expected_choice_0 = {
        "finish_reason": result_0["stopReason"],
        "message": expected_assistant_message,
    }
    assert_choices_in_span(
        span=spans[0],
        expected_choices=[expected_choice_0],
        expect_content=expect_content,
    )

    expected_choice_1 = {
        "finish_reason": result_1["stopReason"],
        "message": {
            "role": result_1["output"]["message"]["role"],
            "content": result_1["output"]["message"]["content"][0]["text"],
        },
    }
    assert_choices_in_span(
        span=spans[1],
        expected_choices=[expected_choice_1],
        expect_content=expect_content,
    )

    metrics = metric_reader.get_metrics_data().resource_metrics
    assert len(metrics) == 1

    metric_data = metrics[0].scope_metrics[0].metrics
    assert_expected_metrics(
        metrics=metric_data,
        request_model=model_id,
        response_model=model_id,
        usage_input_tokens=result_0["usage"]["inputTokens"]
        + result_1["usage"]["inputTokens"],
        usage_output_tokens=result_0["usage"]["outputTokens"]
        + result_1["usage"]["outputTokens"],
    )


@pytest.mark.vcr()
def test_converse_with_content(
    bedrock_client_with_content, claude_model_id: str, span_exporter, metric_reader
):
    _run_and_check_converse(
        bedrock_client=bedrock_client_with_content,
        model_id=claude_model_id,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        expect_content=True,
        stream=False,
    )


@pytest.mark.vcr()
def test_converse_no_content(
    bedrock_client_no_content, claude_model_id: str, span_exporter, metric_reader
):
    _run_and_check_converse(
        bedrock_client=bedrock_client_no_content,
        model_id=claude_model_id,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        expect_content=False,
        stream=False,
    )


@pytest.mark.vcr()
def test_converse_tool_calls_with_content(
    bedrock_client_with_content, claude_model_id: str, span_exporter, metric_reader
):
    _run_and_check_converse_tool_calls(
        bedrock_client=bedrock_client_with_content,
        model_id=claude_model_id,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        expect_content=True,
        stream=False,
    )


@pytest.mark.vcr()
def test_converse_tool_calls_no_content(
    bedrock_client_no_content, claude_model_id: str, span_exporter, metric_reader
):
    _run_and_check_converse_tool_calls(
        bedrock_client=bedrock_client_no_content,
        model_id=claude_model_id,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        expect_content=False,
        stream=False,
    )


@pytest.mark.vcr()
def test_converse_non_existing_model(
    bedrock_client_with_content, span_exporter, metric_reader
):
    model_id = "anthropic.claude-0-0-fake-00000000-v0:0"
    with pytest.raises(Exception):
        bedrock_client_with_content.converse(
            modelId=model_id,
            messages=[{"role": "user", "content": [{"text": "say this is a test"}]}],
        )

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    assert_attributes_in_span(
        span=spans[0],
        span_name="bedrock.converse",
        request_model=model_id,
        error="ValidationException",
    )

    expected_messages = [
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0], expected_messages=expected_messages, expect_content=True
    )

    metrics = metric_reader.get_metrics_data().resource_metrics
    assert len(metrics) == 1

    metric_data = metrics[0].scope_metrics[0].metrics
    assert_expected_metrics(
        metrics=metric_data,
        request_model=model_id,
        error="ValidationException",
    )


@pytest.mark.vcr()
def test_converse_bad_auth(
    instrument_with_content, claude_model_id: str, span_exporter, metric_reader
):
    client = boto3.client(
        "bedrock-runtime",
        aws_access_key_id="test",
        aws_secret_access_key="test",
        aws_session_token="test",
    )
    with pytest.raises(ClientError):
        client.converse(
            modelId=claude_model_id,
            messages=[{"role": "user", "content": [{"text": "say this is a test"}]}],
        )

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    assert_attributes_in_span(
        span=spans[0],
        span_name="bedrock.converse",
        request_model=claude_model_id,
        error="ClientError",
    )

    expected_messages = [
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0], expected_messages=expected_messages, expect_content=True
    )

    metrics = metric_reader.get_metrics_data().resource_metrics
    assert len(metrics) == 1

    metric_data = metrics[0].scope_metrics[0].metrics
    assert_expected_metrics(
        metrics=metric_data,
        request_model=claude_model_id,
        error="ClientError",
    )


@pytest.mark.vcr()
def test_converse_content_blocks(
    bedrock_client_with_content, claude_model_id: str, span_exporter
):
    bedrock_client_with_content.converse(
        modelId=claude_model_id,
        messages=[
            {"role": "user", "content": [{"text": "say this"}, {"text": " is a test"}]}
        ],
    )

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    expected_messages = [
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0], expected_messages=expected_messages, expect_content=True
    )


@pytest.mark.vcr()
def test_converse_unsupported_content_blocks(
    bedrock_client_with_content, claude_model_id: str, span_exporter
):
    bedrock_client_with_content.converse(
        modelId=claude_model_id,
        messages=[
            {
                "role": "user",
                "content": [
                    {"text": "say this"},
                    {"image": {"format": "png", "source": {"bytes": IMAGE_DATA}}},
                    {"text": " is a test"},
                ],
            }
        ],
    )

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    expected_messages = [
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0], expected_messages=expected_messages, expect_content=True
    )


@pytest.mark.vcr()
def test_converse_stream_with_content(
    bedrock_client_with_content, claude_model_id: str, span_exporter, metric_reader
):
    _run_and_check_converse(
        bedrock_client=bedrock_client_with_content,
        model_id=claude_model_id,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        expect_content=True,
        stream=True,
    )


@pytest.mark.vcr()
def test_converse_stream_no_content(
    bedrock_client_no_content, claude_model_id: str, span_exporter, metric_reader
):
    _run_and_check_converse(
        bedrock_client=bedrock_client_no_content,
        model_id=claude_model_id,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        expect_content=False,
        stream=True,
    )


@pytest.mark.vcr()
def test_converse_stream_tool_calls_with_content(
    bedrock_client_with_content, claude_model_id: str, span_exporter, metric_reader
):
    _run_and_check_converse_tool_calls(
        bedrock_client=bedrock_client_with_content,
        model_id=claude_model_id,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        expect_content=True,
        stream=True,
    )


@pytest.mark.vcr()
def test_converse_stream_tool_calls_no_content(
    bedrock_client_no_content, claude_model_id: str, span_exporter, metric_reader
):
    _run_and_check_converse_tool_calls(
        bedrock_client=bedrock_client_no_content,
        model_id=claude_model_id,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        expect_content=False,
        stream=True,
    )


@pytest.mark.vcr()
def test_converse_stream_non_existing_model(
    bedrock_client_with_content, span_exporter, metric_reader
):
    model_id = "anthropic.claude-0-0-fake-00000000-v0:0"
    with pytest.raises(Exception):
        bedrock_client_with_content.converse_stream(
            modelId=model_id,
            messages=[{"role": "user", "content": [{"text": "say this is a test"}]}],
        )

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    assert_attributes_in_span(
        span=spans[0],
        span_name="bedrock.converse_stream",
        request_model=model_id,
        error="ValidationException",
    )

    expected_messages = [
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0], expected_messages=expected_messages, expect_content=True
    )

    metrics = metric_reader.get_metrics_data().resource_metrics
    assert len(metrics) == 1

    metric_data = metrics[0].scope_metrics[0].metrics
    assert_expected_metrics(
        metrics=metric_data,
        request_model=model_id,
        error="ValidationException",
    )


@pytest.mark.vcr()
def test_converse_stream_bad_auth(
    instrument_with_content, claude_model_id: str, span_exporter, metric_reader
):
    client = boto3.client(
        "bedrock-runtime",
        aws_access_key_id="test",
        aws_secret_access_key="test",
        aws_session_token="test",
    )
    with pytest.raises(ClientError):
        client.converse_stream(
            modelId=claude_model_id,
            messages=[{"role": "user", "content": [{"text": "say this is a test"}]}],
        )

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    assert_attributes_in_span(
        span=spans[0],
        span_name="bedrock.converse_stream",
        request_model=claude_model_id,
        error="ClientError",
    )

    expected_messages = [
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0], expected_messages=expected_messages, expect_content=True
    )

    metrics = metric_reader.get_metrics_data().resource_metrics
    assert len(metrics) == 1

    metric_data = metrics[0].scope_metrics[0].metrics
    assert_expected_metrics(
        metrics=metric_data,
        request_model=claude_model_id,
        error="ClientError",
    )


@pytest.mark.vcr()
def test_converse_stream_content_blocks(
    bedrock_client_with_content, claude_model_id: str, span_exporter
):
    result = bedrock_client_with_content.converse_stream(
        modelId=claude_model_id,
        messages=[
            {"role": "user", "content": [{"text": "say this"}, {"text": " is a test"}]}
        ],
    )
    # Consume the stream
    for event in result["stream"]:
        pass

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    expected_messages = [
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0], expected_messages=expected_messages, expect_content=True
    )


@pytest.mark.vcr()
def test_converse_stream_unsupported_content_blocks(
    bedrock_client_with_content, claude_model_id: str, span_exporter
):
    result = bedrock_client_with_content.converse_stream(
        modelId=claude_model_id,
        messages=[
            {
                "role": "user",
                "content": [
                    {"text": "say this"},
                    {"image": {"format": "png", "source": {"bytes": IMAGE_DATA}}},
                    {"text": " is a test"},
                ],
            }
        ],
    )
    # Consume the stream
    for event in result["stream"]:
        pass

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    expected_messages = [
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0], expected_messages=expected_messages, expect_content=True
    )



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=too-many-locals


import pytest
from openai import APIConnectionError, NotFoundError, OpenAI
from opentelemetry.semconv._incubating.attributes import (
    error_attributes as ErrorAttributes,
)
from opentelemetry.semconv._incubating.attributes import (
    gen_ai_attributes as GenAIAttributes,
)
from opentelemetry.semconv._incubating.attributes import (
    server_attributes as ServerAttributes,
)
from opentelemetry.semconv._incubating.metrics import gen_ai_metrics

import llm_tracekit.extended_gen_ai_attributes as ExtendedGenAIAttributes
from tests.openai.utils import (
    assert_all_attributes,
    assert_completion_attributes,
    get_current_weather_tool_definition,
)
from tests.utils import assert_choices_in_span, assert_messages_in_span


@pytest.mark.vcr()
def test_chat_completion_with_content(
    span_exporter, openai_client, instrument_with_content
):
    llm_model_value = "gpt-4o-mini"
    messages_value = [{"role": "user", "content": "Say this is a test"}]

    response = openai_client.chat.completions.create(
        messages=messages_value, model=llm_model_value, stream=False
    )

    spans = span_exporter.get_finished_spans()
    assert_completion_attributes(spans[0], llm_model_value, response)

    user_message = {"role": "user", "content": messages_value[0]["content"]}
    assert_messages_in_span(
        span=spans[0], expected_messages=[user_message], expect_content=True
    )

    choice = {
        "finish_reason": "stop",
        "message": {
            "role": "assistant",
            "content": response.choices[0].message.content,
        },
    }
    assert_choices_in_span(
        span=spans[0], expected_choices=[choice], expect_content=True
    )


@pytest.mark.vcr()
def test_chat_completion_with_content_array(
    span_exporter, openai_client, instrument_with_content
):
    llm_model_value = "gpt-4o-mini"
    messages_value = [
        {
            "role": "user",
            "content": [{"type": "text", "text": "Say this is a test"}],
        }
    ]

    response = openai_client.chat.completions.create(
        messages=messages_value, model=llm_model_value, stream=False
    )

    spans = span_exporter.get_finished_spans()
    assert_completion_attributes(spans[0], llm_model_value, response)

    user_message = {"role": "user"}
    assert_messages_in_span(
        span=spans[0], expected_messages=[user_message], expect_content=True
    )

    choice = {
        "finish_reason": "stop",
        "message": {
            "role": "assistant",
            "content": response.choices[0].message.content,
        },
    }
    assert_choices_in_span(
        span=spans[0], expected_choices=[choice], expect_content=True
    )


@pytest.mark.vcr()
def test_chat_completion_no_content(
    span_exporter, openai_client, instrument_no_content
):
    llm_model_value = "gpt-4o-mini"
    messages_value = [{"role": "user", "content": "Say this is a test"}]

    response = openai_client.chat.completions.create(
        messages=messages_value, model=llm_model_value, stream=False
    )

    spans = span_exporter.get_finished_spans()
    assert_completion_attributes(spans[0], llm_model_value, response)

    assert_messages_in_span(
        span=spans[0], expected_messages=[{"role": "user"}], expect_content=False
    )

    choice = {"finish_reason": "stop", "message": {"role": "assistant"}}
    assert_choices_in_span(
        span=spans[0], expected_choices=[choice], expect_content=False
    )


def test_chat_completion_bad_endpoint(
    span_exporter, metric_reader, instrument_no_content
):
    llm_model_value = "gpt-4o-mini"
    messages_value = [{"role": "user", "content": "Say this is a test"}]

    client = OpenAI(base_url="http://localhost:4242")

    with pytest.raises(APIConnectionError):
        client.chat.completions.create(
            messages=messages_value,
            model=llm_model_value,
            timeout=0.1,
        )

    spans = span_exporter.get_finished_spans()
    assert_all_attributes(spans[0], llm_model_value, server_address="localhost")
    assert 4242 == spans[0].attributes[ServerAttributes.SERVER_PORT]
    assert "APIConnectionError" == spans[0].attributes[ErrorAttributes.ERROR_TYPE]

    metrics = metric_reader.get_metrics_data().resource_metrics
    assert len(metrics) == 1

    metric_data = metrics[0].scope_metrics[0].metrics
    duration_metric = next(
        (
            m
            for m in metric_data
            if m.name == gen_ai_metrics.GEN_AI_CLIENT_OPERATION_DURATION
        ),
        None,
    )
    assert duration_metric is not None
    assert duration_metric.data.data_points[0].sum > 0
    assert (
        duration_metric.data.data_points[0].attributes[ErrorAttributes.ERROR_TYPE]
        == "APIConnectionError"
    )


@pytest.mark.vcr()
def test_chat_completion_404(
    span_exporter, openai_client, metric_reader, instrument_no_content
):
    llm_model_value = "this-model-does-not-exist"
    messages_value = [{"role": "user", "content": "Say this is a test"}]

    with pytest.raises(NotFoundError):
        openai_client.chat.completions.create(
            messages=messages_value,
            model=llm_model_value,
        )

    spans = span_exporter.get_finished_spans()

    assert_all_attributes(spans[0], llm_model_value)
    assert "NotFoundError" == spans[0].attributes[ErrorAttributes.ERROR_TYPE]

    metrics = metric_reader.get_metrics_data().resource_metrics
    assert len(metrics) == 1

    metric_data = metrics[0].scope_metrics[0].metrics
    duration_metric = next(
        (
            m
            for m in metric_data
            if m.name == gen_ai_metrics.GEN_AI_CLIENT_OPERATION_DURATION
        ),
        None,
    )
    assert duration_metric is not None
    assert duration_metric.data.data_points[0].sum > 0
    assert (
        duration_metric.data.data_points[0].attributes[ErrorAttributes.ERROR_TYPE]
        == "NotFoundError"
    )


@pytest.mark.vcr()
def test_chat_completion_extra_params(
    span_exporter, openai_client, instrument_no_content
):
    llm_model_value = "gpt-4o-mini"
    messages_value = [{"role": "user", "content": "Say this is a test"}]

    response = openai_client.chat.completions.create(
        messages=messages_value,
        model=llm_model_value,
        seed=42,
        temperature=0.5,
        max_tokens=50,
        stream=False,
        extra_body={"service_tier": "default"},
        response_format={"type": "text"},
        user="test_user",
    )

    spans = span_exporter.get_finished_spans()
    assert_completion_attributes(spans[0], llm_model_value, response)
    assert spans[0].attributes[GenAIAttributes.GEN_AI_OPENAI_REQUEST_SEED] == 42
    assert spans[0].attributes[GenAIAttributes.GEN_AI_REQUEST_TEMPERATURE] == 0.5
    assert spans[0].attributes[GenAIAttributes.GEN_AI_REQUEST_MAX_TOKENS] == 50
    assert (
        spans[0].attributes[GenAIAttributes.GEN_AI_OPENAI_REQUEST_SERVICE_TIER]
        == "default"
    )
    assert (
        spans[0].attributes[GenAIAttributes.GEN_AI_OPENAI_REQUEST_RESPONSE_FORMAT]
        == "text"
    )
    assert (
        spans[0].attributes[ExtendedGenAIAttributes.GEN_AI_OPENAI_REQUEST_USER]
        == "test_user"
    )


@pytest.mark.vcr()
def test_chat_completion_multiple_choices(
    span_exporter, openai_client, instrument_with_content
):
    llm_model_value = "gpt-4o-mini"
    messages_value = [{"role": "user", "content": "Say this is a test"}]

    response = openai_client.chat.completions.create(
        messages=messages_value, model=llm_model_value, n=2, stream=False
    )

    spans = span_exporter.get_finished_spans()
    assert_completion_attributes(spans[0], llm_model_value, response)

    user_message = {"role": "user", "content": messages_value[0]["content"]}
    assert_messages_in_span(
        span=spans[0], expected_messages=[user_message], expect_content=True
    )

    choices = [
        {
            "finish_reason": "stop",
            "message": {
                "role": "assistant",
                "content": response.choices[0].message.content,
            },
        },
        {
            "finish_reason": "stop",
            "message": {
                "role": "assistant",
                "content": response.choices[1].message.content,
            },
        },
    ]
    assert_choices_in_span(span=spans[0], expected_choices=choices, expect_content=True)


@pytest.mark.vcr()
def test_chat_completion_tool_calls_with_content(
    span_exporter, openai_client, instrument_with_content
):
    chat_completion_tool_call(span_exporter, openai_client, True)


@pytest.mark.vcr()
def test_chat_completion_tool_calls_no_content(
    span_exporter, openai_client, instrument_no_content
):
    chat_completion_tool_call(span_exporter, openai_client, False)


def chat_completion_tool_call(span_exporter, openai_client, expect_content):
    llm_model_value = "gpt-4o-mini"
    messages_value = [
        {"role": "system", "content": "You're a helpful assistant."},
        {
            "role": "user",
            "content": "What's the weather in Seattle and San Francisco today?",
        },
    ]

    response_0 = openai_client.chat.completions.create(
        messages=messages_value,
        model=llm_model_value,
        tool_choice="auto",
        tools=[get_current_weather_tool_definition()],
    )

    # sanity check
    assert "tool_calls" in response_0.choices[0].finish_reason

    # final request
    messages_value.append(
        {
            "role": "assistant",
            "tool_calls": response_0.choices[0].message.to_dict()["tool_calls"],
        }
    )

    tool_call_result_0 = {
        "role": "tool",
        "content": "50 degrees and raining",
        "tool_call_id": response_0.choices[0].message.tool_calls[0].id,
    }
    tool_call_result_1 = {
        "role": "tool",
        "content": "70 degrees and sunny",
        "tool_call_id": response_0.choices[0].message.tool_calls[1].id,
    }

    messages_value.append(tool_call_result_0)
    messages_value.append(tool_call_result_1)

    response_1 = openai_client.chat.completions.create(
        messages=messages_value, model=llm_model_value
    )

    # sanity check
    assert "stop" in response_1.choices[0].finish_reason

    # validate both calls
    spans = span_exporter.get_finished_spans()
    assert len(spans) == 2
    assert_completion_attributes(spans[0], llm_model_value, response_0)
    assert_completion_attributes(spans[1], llm_model_value, response_1)

    # call one
    system_message = {"role": "system", "content": messages_value[0]["content"]}
    user_message = {"role": "user", "content": messages_value[1]["content"]}
    assert_messages_in_span(
        span=spans[0],
        expected_messages=[system_message, user_message],
        expect_content=expect_content,
    )

    function_call_0 = {
        "name": "get_current_weather",
        "arguments": (
            response_0.choices[0]
            .message.tool_calls[0]
            .function.arguments.replace("\n", "")
        ),
    }
    function_call_1 = {
        "name": "get_current_weather",
        "arguments": (
            response_0.choices[0]
            .message.tool_calls[1]
            .function.arguments.replace("\n", "")
        ),
    }

    choice = {
        "finish_reason": "tool_calls",
        "message": {
            "role": "assistant",
            "tool_calls": [
                {
                    "id": response_0.choices[0].message.tool_calls[0].id,
                    "type": "function",
                    "function": function_call_0,
                },
                {
                    "id": response_0.choices[0].message.tool_calls[1].id,
                    "type": "function",
                    "function": function_call_1,
                },
            ],
        },
    }
    assert_choices_in_span(
        span=spans[0], expected_choices=[choice], expect_content=expect_content
    )

    # call two
    assistant_tool_call = {
        "role": "assistant",
        "tool_calls": messages_value[2]["tool_calls"],
    }
    tool_message_0 = {
        "role": "tool",
        "tool_call_id": tool_call_result_0["tool_call_id"],
        "content": tool_call_result_0["content"],
    }
    tool_message_1 = {
        "role": "tool",
        "tool_call_id": tool_call_result_1["tool_call_id"],
        "content": tool_call_result_1["content"],
    }

    assert_messages_in_span(
        span=spans[1],
        expected_messages=[
            system_message,
            user_message,
            assistant_tool_call,
            tool_message_0,
            tool_message_1,
        ],
        expect_content=expect_content,
    )

    choice = {
        "finish_reason": "stop",
        "message": {
            "role": "assistant",
            "content": response_1.choices[0].message.content,
        },
    }

    assert_choices_in_span(
        span=spans[1], expected_choices=[choice], expect_content=expect_content
    )


@pytest.mark.vcr()
def test_chat_completion_streaming(
    span_exporter, openai_client, instrument_with_content
):
    llm_model_value = "gpt-4"
    messages_value = [{"role": "user", "content": "Say this is a test"}]

    kwargs = {
        "model": llm_model_value,
        "messages": messages_value,
        "stream": True,
        "stream_options": {"include_usage": True},
    }

    response_stream_usage = None
    response_stream_model = None
    response_stream_id = None
    response_stream_result = ""
    response = openai_client.chat.completions.create(**kwargs)
    for chunk in response:
        if chunk.choices:
            response_stream_result += chunk.choices[0].delta.content or ""

        # get the last chunk
        if getattr(chunk, "usage", None):
            response_stream_usage = chunk.usage
            response_stream_model = chunk.model
            response_stream_id = chunk.id

    spans = span_exporter.get_finished_spans()
    assert_all_attributes(
        spans[0],
        llm_model_value,
        response_stream_id,
        response_stream_model,
        response_stream_usage.prompt_tokens,
        response_stream_usage.completion_tokens,
    )

    user_message = {"role": "user", "content": "Say this is a test"}
    assert_messages_in_span(
        span=spans[0], expected_messages=[user_message], expect_content=True
    )

    choice = {
        "finish_reason": "stop",
        "message": {"role": "assistant", "content": response_stream_result},
    }
    assert_choices_in_span(
        span=spans[0], expected_choices=[choice], expect_content=True
    )


@pytest.mark.vcr()
def test_chat_completion_streaming_not_complete(
    span_exporter, openai_client, instrument_with_content
):
    llm_model_value = "gpt-4"
    messages_value = [{"role": "user", "content": "Say this is a test"}]

    kwargs = {
        "model": llm_model_value,
        "messages": messages_value,
        "stream": True,
    }

    response_stream_model = None
    response_stream_id = None
    response_stream_result = ""
    response = openai_client.chat.completions.create(**kwargs)
    for idx, chunk in enumerate(response):
        if chunk.choices:
            response_stream_result += chunk.choices[0].delta.content or ""
        if idx == 1:
            # fake a stop
            break

        if chunk.model:
            response_stream_model = chunk.model
        if chunk.id:
            response_stream_id = chunk.id

    response.close()
    spans = span_exporter.get_finished_spans()
    assert_all_attributes(
        spans[0], llm_model_value, response_stream_id, response_stream_model, 0, 0
    )

    user_message = {"role": "user", "content": "Say this is a test"}
    assert_messages_in_span(
        span=spans[0], expected_messages=[user_message], expect_content=True
    )

    choice = {
        "finish_reason": "error",
        "message": {"role": "assistant", "content": response_stream_result},
    }
    assert_choices_in_span(
        span=spans[0], expected_choices=[choice], expect_content=True
    )


@pytest.mark.vcr()
def test_chat_completion_multiple_choices_streaming(
    span_exporter, openai_client, instrument_with_content
):
    llm_model_value = "gpt-4o-mini"
    messages_value = [
        {"role": "system", "content": "You're a helpful assistant."},
        {
            "role": "user",
            "content": "What's the weather in Seattle and San Francisco today?",
        },
    ]

    response_0 = openai_client.chat.completions.create(
        messages=messages_value,
        model=llm_model_value,
        n=2,
        stream=True,
        stream_options={"include_usage": True},
    )

    # two strings for each choice
    response_stream_result = ["", ""]
    finish_reasons = ["", ""]
    for chunk in response_0:
        if chunk.choices:
            for choice in chunk.choices:
                response_stream_result[choice.index] += choice.delta.content or ""
                if choice.finish_reason:
                    finish_reasons[choice.index] = choice.finish_reason

        # get the last chunk
        if getattr(chunk, "usage", None):
            response_stream_usage = chunk.usage
            response_stream_model = chunk.model
            response_stream_id = chunk.id

    # sanity check
    assert "stop" == finish_reasons[0]

    spans = span_exporter.get_finished_spans()
    assert_all_attributes(
        spans[0],
        llm_model_value,
        response_stream_id,
        response_stream_model,
        response_stream_usage.prompt_tokens,
        response_stream_usage.completion_tokens,
    )

    messages = [
        {"role": "system", "content": messages_value[0]["content"]},
        {
            "role": "user",
            "content": "What's the weather in Seattle and San Francisco today?",
        },
    ]
    assert_messages_in_span(
        span=spans[0], expected_messages=messages, expect_content=True
    )

    choices = [
        {
            "finish_reason": "stop",
            "message": {
                "role": "assistant",
                "content": "".join(response_stream_result[0]),
            },
        },
        {
            "finish_reason": "stop",
            "message": {
                "role": "assistant",
                "content": "".join(response_stream_result[1]),
            },
        },
    ]
    assert_choices_in_span(span=spans[0], expected_choices=choices, expect_content=True)


@pytest.mark.vcr()
def test_chat_completion_multiple_tools_streaming_with_content(
    span_exporter, openai_client, instrument_with_content
):
    chat_completion_multiple_tools_streaming(span_exporter, openai_client, True)


@pytest.mark.vcr()
def test_chat_completion_multiple_tools_streaming_no_content(
    span_exporter, openai_client, instrument_no_content
):
    chat_completion_multiple_tools_streaming(span_exporter, openai_client, False)


@pytest.mark.vcr()
def test_chat_completion_with_content_span_unsampled(
    span_exporter,
    openai_client,
    instrument_with_content_unsampled,
):
    llm_model_value = "gpt-4o-mini"
    messages_value = [{"role": "user", "content": "Say this is a test"}]

    openai_client.chat.completions.create(
        messages=messages_value, model=llm_model_value, stream=False
    )

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 0


def chat_completion_multiple_tools_streaming(
    span_exporter, openai_client, expect_content
):
    llm_model_value = "gpt-4o-mini"
    messages_value = [
        {"role": "system", "content": "You're a helpful assistant."},
        {
            "role": "user",
            "content": "What's the weather in Seattle and San Francisco today?",
        },
    ]

    response = openai_client.chat.completions.create(
        messages=messages_value,
        model=llm_model_value,
        tool_choice="auto",
        tools=[get_current_weather_tool_definition()],
        stream=True,
        stream_options={"include_usage": True},
    )

    finish_reason = None
    # two tools
    tool_names = ["", ""]
    tool_call_ids = ["", ""]
    tool_args = ["", ""]
    for chunk in response:
        if chunk.choices:
            if chunk.choices[0].finish_reason:
                finish_reason = chunk.choices[0].finish_reason
            for tool_call in chunk.choices[0].delta.tool_calls or []:
                t_idx = tool_call.index
                if tool_call.id:
                    tool_call_ids[t_idx] = tool_call.id
                if tool_call.function:
                    if tool_call.function.arguments:
                        tool_args[t_idx] += tool_call.function.arguments
                    if tool_call.function.name:
                        tool_names[t_idx] = tool_call.function.name

        # get the last chunk
        if getattr(chunk, "usage", None):
            response_stream_usage = chunk.usage
            response_stream_model = chunk.model
            response_stream_id = chunk.id

    # sanity check
    assert "tool_calls" == finish_reason

    spans = span_exporter.get_finished_spans()
    assert_all_attributes(
        spans[0],
        llm_model_value,
        response_stream_id,
        response_stream_model,
        response_stream_usage.prompt_tokens,
        response_stream_usage.completion_tokens,
    )

    system_message = {"role": "system", "content": messages_value[0]["content"]}
    user_message = {
        "role": "user",
        "content": "What's the weather in Seattle and San Francisco today?",
    }

    assert_messages_in_span(
        span=spans[0],
        expected_messages=[system_message, user_message],
        expect_content=expect_content,
    )

    tool_call_0 = {
        "id": tool_call_ids[0],
        "type": "function",
        "function": {
            "name": tool_names[0],
            "arguments": tool_args[0].replace("\n", ""),
        },
    }
    tool_call_1 = {
        "id": tool_call_ids[1],
        "type": "function",
        "function": {
            "name": tool_names[1],
            "arguments": tool_args[1].replace("\n", ""),
        },
    }

    choice = {
        "finish_reason": "tool_calls",
        "message": {
            "role": "assistant",
            "tool_calls": [tool_call_0, tool_call_1],
        },
    }
    assert_choices_in_span(
        span=spans[0], expected_choices=[choice], expect_content=expect_content
    )



# Copyright Coralogix Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial, wraps
from io import BytesIO
from timeit import default_timer
from typing import Callable, Optional

from botocore.eventstream import EventStream
from botocore.response import StreamingBody
from opentelemetry.trace import Span, SpanKind, Tracer

from llm_tracekit.bedrock.converse import (
    ConverseStreamWrapper,
    generate_attributes_from_converse_input,
    record_converse_result_attributes,
)
from llm_tracekit.bedrock.invoke_agent import (
    InvokeAgentStreamWrapper,
    generate_attributes_from_invoke_agent_input,
    record_invoke_agent_result_attributes,
)
from llm_tracekit.bedrock.invoke_model import (
    InvokeModelWithResponseStreamWrapper,
    generate_attributes_from_invoke_input,
    record_invoke_model_result_attributes,
)
from llm_tracekit.bedrock.utils import record_metrics
from llm_tracekit.instrumentation_utils import handle_span_exception
from llm_tracekit.instruments import Instruments


def _handle_error(
    error: Exception,
    span: Span,
    start_time: float,
    instruments: Instruments,
    model: Optional[str] = None,
):
    duration = max((default_timer() - start_time), 0)
    handle_span_exception(span, error)
    record_metrics(
        instruments=instruments,
        duration=duration,
        request_model=model,
        error_type=type(error).__qualname__,
    )


def invoke_model_wrapper(
    original_function: Callable,
    tracer: Tracer,
    instruments: Instruments,
    capture_content: bool,
):
    @wraps(original_function)
    def wrapper(*args, **kwargs):
        model = kwargs.get("modelId")
        span_attributes = generate_attributes_from_invoke_input(
            kwargs=kwargs, capture_content=capture_content
        )
        with tracer.start_as_current_span(
            name="bedrock.invoke_model",
            kind=SpanKind.CLIENT,
            attributes=span_attributes,
            end_on_exit=False,
        ) as span:
            start_time = default_timer()
            try:
                result = original_function(*args, **kwargs)
                body = result.get("body")
                if body is not None:
                    body_content = body.read()
                    # The response body is a stream, and reading the stream consumes it, so we have to recreate
                    # it to keep the original response usable
                    result["body"] = StreamingBody(
                        BytesIO(body_content), len(body_content)
                    )

                    record_invoke_model_result_attributes(
                        result_body=body_content,
                        span=span,
                        start_time=start_time,
                        instruments=instruments,
                        capture_content=capture_content,
                        model_id=model,
                    )

                return result
            except Exception as error:
                _handle_error(
                    error=error,
                    span=span,
                    start_time=start_time,
                    instruments=instruments,
                    model=model,
                )
                raise

    return wrapper


def invoke_model_with_response_stream_wrapper(
    original_function: Callable,
    tracer: Tracer,
    instruments: Instruments,
    capture_content: bool,
):
    @wraps(original_function)
    def wrapper(*args, **kwargs):
        model = kwargs.get("modelId")
        span_attributes = generate_attributes_from_invoke_input(
            kwargs=kwargs, capture_content=capture_content
        )
        with tracer.start_as_current_span(
            name="bedrock.invoke_model_with_response_stream",
            kind=SpanKind.CLIENT,
            attributes=span_attributes,
            end_on_exit=False,
        ) as span:
            start_time = default_timer()
            try:
                result = original_function(*args, **kwargs)
                if "body" in result and isinstance(result["body"], EventStream):
                    result["body"] = InvokeModelWithResponseStreamWrapper(
                        stream=result["body"],
                        stream_done_callback=partial(
                            record_invoke_model_result_attributes,
                            span=span,
                            start_time=start_time,
                            instruments=instruments,
                            capture_content=capture_content,
                            model_id=model,
                        ),
                        stream_error_callback=partial(
                            _handle_error,
                            span=span,
                            start_time=start_time,
                            instruments=instruments,
                            model=model,
                        ),
                        model_id=model,
                    )

                return result
            except Exception as error:
                _handle_error(
                    error=error,
                    span=span,
                    start_time=start_time,
                    instruments=instruments,
                    model=model,
                )
                raise

    return wrapper


def converse_wrapper(
    original_function: Callable,
    tracer: Tracer,
    instruments: Instruments,
    capture_content: bool,
):
    @wraps(original_function)
    def wrapper(*args, **kwargs):
        model = kwargs.get("modelId")
        span_attributes = generate_attributes_from_converse_input(
            kwargs=kwargs, capture_content=capture_content
        )

        with tracer.start_as_current_span(
            name="bedrock.converse",
            kind=SpanKind.CLIENT,
            attributes=span_attributes,
            end_on_exit=False,
        ) as span:
            start_time = default_timer()
            try:
                result = original_function(*args, **kwargs)
                record_converse_result_attributes(
                    result=result,
                    span=span,
                    start_time=start_time,
                    instruments=instruments,
                    capture_content=capture_content,
                    model=model,
                )
                return result
            except Exception as error:
                _handle_error(
                    error=error,
                    span=span,
                    start_time=start_time,
                    instruments=instruments,
                    model=model,
                )
                raise

    return wrapper


def converse_stream_wrapper(
    original_function: Callable,
    tracer: Tracer,
    instruments: Instruments,
    capture_content: bool,
):
    @wraps(original_function)
    def wrapper(*args, **kwargs):
        model = kwargs.get("modelId")
        span_attributes = generate_attributes_from_converse_input(
            kwargs=kwargs, capture_content=capture_content
        )

        with tracer.start_as_current_span(
            name="bedrock.converse_stream",
            kind=SpanKind.CLIENT,
            attributes=span_attributes,
            end_on_exit=False,
        ) as span:
            start_time = default_timer()
            try:
                result = original_function(*args, **kwargs)
                if "stream" in result and isinstance(result["stream"], EventStream):
                    result["stream"] = ConverseStreamWrapper(
                        stream=result["stream"],
                        stream_done_callback=partial(
                            record_converse_result_attributes,
                            span=span,
                            start_time=start_time,
                            instruments=instruments,
                            capture_content=capture_content,
                            model=model,
                        ),
                        stream_error_callback=partial(
                            _handle_error,
                            span=span,
                            start_time=start_time,
                            instruments=instruments,
                            model=model,
                        ),
                    )

                return result
            except Exception as error:
                _handle_error(
                    error=error,
                    span=span,
                    start_time=start_time,
                    instruments=instruments,
                    model=model,
                )
                raise

    return wrapper


def invoke_agent_wrapper(
    original_function: Callable,
    tracer: Tracer,
    instruments: Instruments,
    capture_content: bool,
):
    @wraps(original_function)
    def wrapper(*args, **kwargs):
        span_attributes = generate_attributes_from_invoke_agent_input(
            kwargs=kwargs, capture_content=capture_content
        )
        with tracer.start_as_current_span(
            name="bedrock.invoke_agent",
            kind=SpanKind.CLIENT,
            attributes=span_attributes,
            end_on_exit=False,
        ) as span:
            start_time = default_timer()
            try:
                result = original_function(*args, **kwargs)
                if "completion" in result:
                    result["completion"] = InvokeAgentStreamWrapper(
                        stream=result["completion"],
                        stream_done_callback=partial(
                            record_invoke_agent_result_attributes,
                            span=span,
                            start_time=start_time,
                            instruments=instruments,
                            capture_content=capture_content,
                        ),
                        stream_error_callback=partial(
                            _handle_error,
                            span=span,
                            start_time=start_time,
                            instruments=instruments,
                        ),
                    )

                return result
            except Exception as error:
                _handle_error(
                    error=error,
                    span=span,
                    start_time=start_time,
                    instruments=instruments,
                )
                raise

    return wrapper


def create_client_wrapper(
    tracer: Tracer,
    instruments: Instruments,
    capture_content: bool,
):
    def traced_method(wrapped, instance, args, kwargs):
        service_name = kwargs.get("service_name")
        client = wrapped(*args, **kwargs)
        if service_name == "bedrock-runtime":
            client.invoke_model = invoke_model_wrapper(
                original_function=client.invoke_model,
                tracer=tracer,
                instruments=instruments,
                capture_content=capture_content,
            )
            client.invoke_model_with_response_stream = (
                invoke_model_with_response_stream_wrapper(
                    original_function=client.invoke_model_with_response_stream,
                    tracer=tracer,
                    instruments=instruments,
                    capture_content=capture_content,
                )
            )
            client.converse = converse_wrapper(
                original_function=client.converse,
                tracer=tracer,
                instruments=instruments,
                capture_content=capture_content,
            )
            client.converse_stream = converse_stream_wrapper(
                original_function=client.converse_stream,
                tracer=tracer,
                instruments=instruments,
                capture_content=capture_content,
            )
        elif service_name == "bedrock-agent-runtime":
            client.invoke_agent = invoke_agent_wrapper(
                original_function=client.invoke_agent,
                tracer=tracer,
                instruments=instruments,
                capture_content=capture_content,
            )

        return client

    return traced_method



# Copyright Coralogix Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from opentelemetry.semconv.attributes import (
    error_attributes as ErrorAttributes,
)
from opentelemetry.trace.status import Status, StatusCode

OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT = (
    "OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT"
)


def is_content_enabled() -> bool:
    """Checks if capturing message content is enabled."""
    capture_content = os.environ.get(
        OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT, "false"
    )

    return capture_content.lower() == "true"


def enable_capture_content():
    """Enables capturing message content."""
    os.environ[OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT] = "true"


def handle_span_exception(span, error):
    span.set_status(Status(StatusCode.ERROR, str(error)))
    if span.is_recording():
        span.set_attribute(ErrorAttributes.ERROR_TYPE, type(error).__qualname__)
    span.end()
